import json
from enum import Enum


def to_json(obj):
    return json.dumps(obj, default=lambda o: o.__dict__, sort_keys=True)

class AutoName(Enum):
    """
    https://docs.python.org/3/library/enum.html#using-automatic-values
    """
    def _generate_next_value_(name, start, count, last_values):
        return name

class AutoNameLower(Enum):
    """
    https://docs.python.org/3/library/enum.html#using-automatic-values
    """
    def _generate_next_value_(name, start, count, last_values):
        return name.lower()



from typing import Optional, List

from pydantic import BaseModel, HttpUrl


class Persona(BaseModel):
    name: str
    profile_picture_url: HttpUrl

class PersonaWithId(Persona):
    id: str

class PersonaResponse(BaseModel):
    id: str

class Cursors(BaseModel):
    before: str
    after: str

class Paging(BaseModel):
    cursors: Cursors

class AllPerosnasResponse(BaseModel):
    data: List[PersonaWithId]
    paging: Paging


class PersonaDeleteResponse(BaseModel):
    success: bool


from typing import Optional, List, Literal

from pydantic import BaseModel, HttpUrl, conlist

from songmam.facebook.messaging.templates.button import AllButtonTypes


class GenericElement(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/generic#elements
    """
    title: str
    subtitle: Optional[str]
    image_url: Optional[HttpUrl]
    default_action: Optional[AllButtonTypes]  # link to url if press on the image that showed
    buttons: Optional[conlist(AllButtonTypes, min_items=1, max_items=3)]


class PayloadGeneric(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/generic
    """
    template_type: str = "generic"
    image_aspect_ratio: Optional[Literal["horizontal", "square"]] = None
    elements: List[GenericElement]



from typing import Literal

from pydantic import BaseModel

from songmam.facebook.entries.base import MessagingWithTimestamp


class AccountLink(BaseModel):
    status: Literal["linked", "unlinked"]
    authorization_code: str

class AccountLinkEntry(MessagingWithTimestamp):
    account_linking: AccountLink

# {
#   "recipient":{
#     "id":"USER_ID"
#   },
#   "recipient":{
#     "id":"PAGE_ID"
#   },
#   "timestamp":1234567890,
#   "account_linking":{
#     "status":"linked",
#     "authorization_code":"PASS_THROUGH_AUTHORIZATION_CODE"
#   }
# }


from pydantic import BaseModel

from songmam.facebook.webhook import Webhook

class StandbyEntry(BaseModel):
    id: str
    time: int
    standby: List[]

class StandbyObject(Webhook):
    entry: StandbyEntry

# {
#   "object":"page",
#   "entry":[
#     {
#       "id":"<PAGE_ID>",
#       "time":1458692752478,
#       "standby":[
#         {
#           "recipient":{
#             "id":"<USER_ID>"
#           },
#           "recipient":{
#             "id":"<PAGE_ID>"
#           },
#
#           ...
#         }
#       ]
#     }
#   ]
# }


# https://developers.facebook.com/docs/messenger-platform/reference/messenger-profile-api/greeting
from typing import List

from pydantic import validator

from songmam.facebook.messaging.locale import ThingWithLocale


class GreetingPerLocale(ThingWithLocale):
    text: str  # Text Could be Personalized e.g. "Hello {{user_first_name}}!"

    @validator('text')
    def char_limit_to_160(cls, value):
        if len(value) > 160:
            raise ValueError("Must be in UTF-8. 160 character limit.")
        return value




from pydantic import BaseModel, Field

from songmam.facebook import ThingWithId


class PolicyEnforcement(BaseModel):
    action: str
    reason: str

class PolicyEnforcementEntry(BaseModel):
    recipitent: ThingWithId
    timestamp: int
    policy_enforcement: PolicyEnforcement = Field(None, alias="policy-enforcement")

# {
#   "recipient": {
#     "id": "PAGE_ID"
#   },
#   "timestamp": 1458692752478,
#   "policy-enforcement": {
#     "action": "block",
#     "reason": "The bot violated our Platform Policies (https://developers.facebook.com/policy/#messengerplatform). Common violations include sending out excessive spammy messages or being non-functional."
#   }
# }


from typing import Optional

from starlette.responses import PlainTextResponse, Response
from starlette.types import ASGIApp, Receive, Scope, Send
from starlette.requests import Request


class VerificationMiddleware:
    """Verify middleware that follows the step 4: Add webhook verification
    https://developers.facebook.com/docs/messenger-platform/getting-started/webhook-setup
    """
    def __init__(self, app: ASGIApp, verify_token: str, route_webhook: Optional[str] = None) -> None:
        self.app = app
        self.verify_token = verify_token
        self.route_webhook = route_webhook

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "http":
            request = Request(scope, receive)

            if request.method == 'GET':
                query_params = request.query_params
                criteria = {'hub.mode', 'hub.verify_token', 'hub.challenge'}
                if set(query_params.keys()).issuperset(criteria):
                    mode = query_params.get('hub.mode')
                    test_token = query_params.get('hub.verify_token')
                    challenge = query_params.get('hub.challenge')
                    if mode and test_token:
                        if mode == 'subscribe' and test_token == self.verify_token:
                            response = PlainTextResponse(challenge)
                        else:
                            response = Response("", status_code=403)

                        await response(scope, receive, send)
                        return
            await self.app(scope, receive, send)



from typing import Optional

from pydantic import BaseModel

from songmam.facebook.entries.base import MessagingWithTimestamp
from songmam.facebook.entries.messages import MessageEntry, Message as Message_


class Message(Message_):
    is_echo: bool
    app_id: str
    metadata: Optional[str]

class EchoEntry(MessageEntry):
    message: Message

# {
#   "recipient":{
#     "id":"<PSID>"
#   },
#   "recipient":{
#     "id":"<USER_ID>"
#   },
#   "timestamp":1457764197627,
#   "message":{
#     "is_echo":true,
#     "app_id":1517776481860111,
#     "metadata": "<DEVELOPER_DEFINED_METADATA_STRING>",
#     "mid":"mid.1457764197618:41d102a3e1ae206a38",
#     ...
#   }
# }


# https://developers.facebook.com/docs/messenger-platform/send-messages#send_api_basics
from pydantic import BaseModel
from typing import Optional, Literal, List

from songmam.facebook import ThingWithId
from songmam.facebook.messaging.messaging_type import MessagingType


class SendResponse(BaseModel):
    recipient_id: Optional[str]
    message_id: Optional[str]


class SendRecipient(ThingWithId):
    user_ref: Optional[str]
    post_id: Optional[str]
    comment_id: Optional[str]


class MessageAttachmentPayload(BaseModel):
    # Payload of attachment, can either be
    # a Template Payload
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates#payload
    """
    template_type: Optional[str]
    # ...: Mixed ???
    # or
    # a File Attachment Payload
    """
    https://developers.facebook.com/docs/messenger-platform/reference/attachment-upload-api#payload
    """
    url: Optional[str]
    is_reusable: Optional[bool]


class MessageAttachment(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/send-api/#attachment
    """
    type: Optional[str]
    payload: Optional[MessageAttachmentPayload]


class SendMessage(BaseModel):
    text: Optional[str]
    attachment: Optional[MessageAttachment]
    quick_replies: Optional[List[str]] # Array<quick_reply>
    metadata: Optional[str]


class RequestUriEntry(BaseModel):
    messaging_type: Optional[MessagingType]
    recipient: Optional[SendRecipient]
    message: Optional[SendMessage]
    sender_action: Optional[Literal["typing_on", "typing_off", "mark_seen"]] # confused by description
    notification_type: Optional[str]
    tag: Optional[str]



import json

from songmam.facebook.entries.base import BaseMessaging
from songmam.facebook.entries.deliveries import DeliveriesEntry
from songmam.facebook.entries.messages import MessageEntry
from songmam.facebook.entries.postback import PostbackEntry
from songmam.facebook.entries.referral import ReferralEntry


class Event:
    entry: MessageEntry

    def __init__(self, entry):
        self.entry = entry
        self.sender = self.entry.theMessaging.sender
        self.recipient = self.entry.theMessaging.recipient

    def __str__(self):
        # TODO: some infomative
        return json.dumps(self.__class__.__name__)


class MessageEvent(Event):
    entry: MessageEntry

    @property
    def is_quick_reply(self):
        if self.entry.theMessaging.message.quick_reply:
            return True
        else:
            return False

    def __init__(self, entry):
        super(MessageEvent, self).__init__(entry)

        self.name = 'message'
        # self.message = self.entry.theMessaging.message
        self.text = self.entry.theMessaging.message.text
        self.message_id = self.entry.theMessaging.message.mid
        self.quick_reply = self.entry.theMessaging.message.quick_reply
        self.reply_to = self.entry.theMessaging.message.reply_to
        self.attachments = self.entry.theMessaging.message.attachments


class PostBackEvent(Event):
    entry: PostbackEntry

    def __init__(self, entry: PostbackEntry):
        super(PostBackEvent, self).__init__(entry)

        self.title = self.entry.theMessaging.postback.title
        self.payload = self.entry.theMessaging.postback.payload
        # self.referal = self.entry.postback.referral



class DeliveriesEvent(Event):
    entry: DeliveriesEntry

    def __init__(self, entry: DeliveriesEntry):
        super(DeliveriesEvent, self).__init__(entry)

        self.mids = self.entry.theMessaging.delivery.mids
        self.watermark = self.entry.theMessaging.delivery.watermark


class EchoEvent(Event):
    def __init__(self, message, **kwargs):
        super(EchoEvent, self).__init__(**kwargs)

        self.name = 'echo'
        self.message = message

    @property
    def mid(self):
        return self.message.get('mid')

    @property
    def app_id(self):
        return self.message.get('app_id')

    @property
    def metadata(self):
        return self.message.get('metadata')

    @property
    def text(self):
        return self.message.get('text')

    @property
    def attachments(self):
        return self.message.get('attachments')


class ReadEvent(Event):
    def __init__(self, read, **kwargs):
        super(ReadEvent, self).__init__(**kwargs)

        self.name = 'read'
        self.read = read

    @property
    def seq(self):
        return self.read.get('seq')

    @property
    def watermark(self):
        return self.read.get('watermark')


class AccountLinkingEvent(Event):
    def __init__(self, account_linking, **kwargs):
        super(AccountLinkingEvent, self).__init__(**kwargs)

        self.name = 'account_linking'
        self.account_linking = account_linking

    @property
    def status(self):
        return self.account_linking.get('status')

    @property
    def is_linked(self):
        return self.status == 'linked'

    @property
    def authorization_code(self):
        return self.account_linking.get('authorization_code')


class GamePlayEvent(Event):
    def __init__(self, game_play, **kwargs):
        super(GamePlayEvent, self).__init__(**kwargs)

        self.name = 'game_play'
        self.game_play = game_play

    @property
    def game_id(self):
        return self.game_play.get('game_id')

    @property
    def player_id(self):
        return self.game_play.get('player_id')

    @property
    def context_type(self):
        return self.game_play.get('context_type')

    @property
    def context_id(self):
        return self.game_play.get('context_id')

    @property
    def score(self):
        return self.game_play.get('score')

    @property
    def payload(self):
        return self.game_play.get('payload')


class PassThreadEvent(Event):
    def __init__(self, pass_thread_control, **kwargs):
        super(PassThreadEvent, self).__init__(**kwargs)

        self.name = 'pass_thread_control'
        self.pass_thread_control = pass_thread_control

    @property
    def new_owner_app_id(self):
        return self.pass_thread_control.get('new_owner_app_id')

    @property
    def metadata(self):
        return self.pass_thread_control.get('metadata')


class TakeThreadEvent(Event):
    def __init__(self, take_thread_control, **kwargs):
        super(TakeThreadEvent, self).__init__(**kwargs)

        self.name = 'take_thread_control'
        self.take_thread_control = take_thread_control

    @property
    def previous_owner_app_id(self):
        return self.take_thread_control.get('previous_owner_app_id')

    @property
    def metadata(self):
        return self.take_thread_control.get('metadata')


class RequestThreadEvent(Event):
    def __init__(self, request_thread_control, **kwargs):
        super(RequestThreadEvent, self).__init__(**kwargs)

        self.name = 'request_thread_control'
        self.request_thread_control = request_thread_control

    @property
    def requested_owner_app_id(self):
        return self.request_thread_control.get('requested_owner_app_id')

    @property
    def metadata(self):
        return self.request_thread_control.get('metadata')


class AppRoleEvent(Event):
    def __init__(self, app_roles, **kwargs):
        super(AppRoleEvent, self).__init__(**kwargs)

        self.name = 'app_roles'
        self.app_roles = app_roles


class OptinEvent(Event):
    def __init__(self, optin, **kwargs):
        super(OptinEvent, self).__init__(**kwargs)

        self.name = 'optin'
        self.optin = optin

    @property
    def ref(self):
        return self.optin.get('ref')

    @property
    def user_ref(self):
        return self.optin.get('user_ref')


class PolicyEnforcementEvent(Event):
    def __init__(self, policy_enforcement, **kwargs):
        super(PolicyEnforcementEvent, self).__init__(**kwargs)

        self.name = 'policy_enforcement'
        self.policy_enforcement = policy_enforcement

    @property
    def action(self):
        return self.policy_enforcement.get('action')

    @property
    def reason(self):
        return self.policy_enforcement.get('reason')


class ReferralEvent(Event):
    entry: ReferralEntry

    def __init__(self, entry):
        super(ReferralEvent, self).__init__(entry)
        self.referral = self.entry.referral

    # @property
    # def source(self):
    #     return self.referral.get('source')
    #
    # @property
    # def type(self):
    #     return self.referral.get('type')
    #
    # @property
    # def ref(self):
    #     return self.referral.get('ref')
    #
    # @property
    # def referer_uri(self):
    #     return self.referral.get('referer_uri')


class CheckOutUpdateEvent(Event):  # beta
    def __init__(self, checkout_update, **kwargs):
        super(CheckOutUpdateEvent, self).__init__(**kwargs)

        self.name = 'checkout_update'
        self.checkout_update = checkout_update

    @property
    def payload(self):
        return self.checkout_update.get('payload')

    @property
    def shipping_address(self):
        return self.checkout_update.get('shipping_address')


class PaymentEvent(Event):  # beta
    def __init__(self, payment, **kwargs):
        super(PaymentEvent, self).__init__(**kwargs)

        self.name = 'payment'
        self.payment = payment

    @property
    def payload(self):
        return self.payment.get('payload')

    @property
    def requested_user_info(self):
        return self.payment.get('requested_user_info')

    @property
    def payment_credential(self):
        return self.payment.get('payment_credential')

    @property
    def amount(self):
        return self.payment.get('amount')

    @property
    def shipping_option_id(self):
        return self.payment.get('shipping_option_id')


class StandByEvent(Event):
    def __init__(self, standby, **kwargs):
        super(StandByEvent, self).__init__(**kwargs)

        self.name = 'standby'
        self.standby = standby



from typing import Literal, Union, List, Optional, Type, Any

from pydantic import BaseModel, root_validator, conlist

from songmam.facebook.entries.message.attachment import Attachment as Attachment_
from songmam.facebook.messaging.quick_replies import QuickReply
from songmam.facebook.messaging.templates.button import AllButtonTypes, PayloadButtonTemplate
from songmam.facebook.messaging.templates.generic import PayloadGeneric
from songmam.facebook.messaging.templates.media import PayloadMedia
from songmam.facebook.messaging.templates.receipt import ReceiptElements, Address, Summary, Adjustments, PayloadReceipt
from songmam.facebook.messaging.templates.button import AllButtonTypes


class TemplateAttachment(Attachment_):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-flight-update#attachment
    """
    type: Literal['audio', 'file', 'image', 'location', 'video', 'fallback', "template"] = "template"
    payload: Union[PayloadReceipt, PayloadGeneric, PayloadMedia, PayloadButtonTemplate]


class Message(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-flight-update#message
    https://developers.facebook.com/docs/messenger-platform/reference/send-api
    """
    text: Optional[str]
    attachment: Optional[TemplateAttachment]
    quick_replies: Optional[List[QuickReply]]
    metadata: Optional[str]

    # TODO: Fix this validator
    # @root_validator
    # def text_or_attachment_must_be_set(cls, values):
    #     text, attachment = values.get('text'), values.get('attachment')
    #     counter = 0
    #     if text:
    #         counter += 1
    #     if attachment:
    #         counter += 1
    # TODO: Fix this validator : This line will caused the incomplete initialiazation to throw error. m=Message(), m.text="cat"
    #     if counter == 0:
    #         raise ValueError("text or attachment must be set.")
    #     return values



# https://developers.facebook.com/docs/messenger-platform/send-messages/persistent-menu
from typing import List, Union

from pydantic import BaseModel, conlist

from songmam.facebook.messaging.templates.button import BaseButton, PostbackButton, URLButton
from songmam.facebook.messaging.locale import ThingWithLocale


class MenuPerLocale(ThingWithLocale):
    composer_input_disabled: bool = False
    call_to_actions: conlist(Union[URLButton, PostbackButton], min_items=1, max_items=3)


class PersistentMenu(BaseModel):
    persistent_menu: List[MenuPerLocale]
    
class UserPersistentMenu(PersistentMenu):
    psid: str




from typing import List

from pydantic import BaseModel, conlist

from songmam.facebook.entries.base import MessagingWithTimestamp
from songmam.facebook.entries.messages import MessageEntry


class Delivery(BaseModel):
    mids: List[str]
    watermark: int


class DeliveriesMessaging(MessagingWithTimestamp):
    delivery: Delivery

class DeliveriesEntry(MessageEntry):
    messaging: conlist(DeliveriesMessaging, min_items=1, max_items=1)

"""
https://developers.facebook.com/docs/messenger-platform/reference/webhook-events/message-deliveries
"""



from pydantic import BaseModel

from songmam.facebook.entries.base import MessagingWithTimestamp


class PassThreadControl(BaseModel):
    new_owner_app_id: str
    metadata: str


class HandoversEntry(MessagingWithTimestamp):
    pass_thread_control: PassThreadControl


# {
#   "recipient":{
#     "id":"<PSID>"
#   },
#   "recipient":{
#     "id":"<PAGE_ID>"
#   },
#   "timestamp":1458692752478,
#   "pass_thread_control":{
#     "new_owner_app_id":"123456789",
#     "metadata":"Additional content that the caller wants to set"
#   }
# }


from typing import Optional, Any

from pydantic import BaseModel, HttpUrl


class UserProfile(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/identity/user-profile
    """
    id: str
    first_name: str
    last_name: str
    profile_pic: HttpUrl
    locale: Optional[str] # TODO: to be changed to `literal`
    timezone: Optional[int]
    gender: Optional[str] # TODO: to be changed to `literal`


from pydantic import BaseModel

from songmam.facebook.entries.base import MessagingWithTimestamp
from songmam.facebook import ThingWithId


class Reading(BaseModel):
    watermark: int

class ReadingEntry(MessagingWithTimestamp):
    read: Reading





from typing import Literal, Optional, Union

from pydantic import BaseModel, HttpUrl


class QuickReply(BaseModel):
    content_type: Literal["text", "location", "user_phone_number", "user_email"] = 'text'
    title: Optional[str]  # Required if content_type is 'text'
    payload: Optional[Union[str, int]]  # "payload":"<POSTBACK_PAYLOAD>"
    image_url: Optional[HttpUrl]  # Required if title is an empty string.






# __version__ = '2.5.0'

# from .page import *
from .middleware import VerificationMiddleware
from songmam.facebook.webhook import Webhook
from .facebook.entries.message import attachment as Attachment

# from songmam.middleware import VerificationMiddleware
# from songmam import VerificationMiddleware


import asyncio
import json
import re
import hmac
import hashlib
from typing import Union, Optional, Literal, Set, List, Type, Awaitable

import httpx
import requests
from cacheout import Cache
from fastapi import FastAPI, Request
from furl import furl
from loguru import logger
from pydantic import HttpUrl
from songmam.facebook.messaging.quick_replies import QuickReply
from songmam.facebook.messaging.templates import Message, AllButtonTypes, TemplateAttachment, PayloadButtonTemplate

from .api.events import MessageEvent, PostBackEvent, ReferralEvent, DeliveriesEvent
from .facebook import ThingWithId
from .facebook.entries.deliveries import DeliveriesEntry
from .facebook.entries.echo import EchoEntry
from .facebook.entries.messages import MessageEntry, Sender
from .facebook.entries.postback import PostbackEntry
from .facebook.entries.referral import ReferralEntry
from .facebook.messaging.message_tags import MessageTag
from .facebook.messaging.messaging_type import MessagingType
from .facebook.messaging.notification_type import NotificationType
from .facebook.messaging.payload import CompletePayload, SenderActionPayload
from songmam.facebook.messenger_profile.persistent_menu import UserPersistentMenu, MenuPerLocale
from .facebook.messaging.sender_action import SenderAction
from .facebook.messaging.templates.generic import GenericElement, PayloadGeneric
from .facebook.messaging.templates.media import MediaElement, PayloadMedia
from .facebook.messenger_profile import MessengerProfileProperty, MessengerProfile, GreetingPerLocale
from .facebook.page import Me
from .facebook.persona import Persona, PersonaWithId, PersonaResponse, AllPerosnasResponse, PersonaDeleteResponse
from .facebook.send import SendResponse, SendRecipient
from .facebook.user_profile import UserProfile
from songmam.facebook.webhook import Webhook

# See https://developers.facebook.com/docs/graph-api/changelog
SUPPORTED_API_VERS = Literal[
    "v7.0"  # May 5, 2020
]


class Page:
    access_token: str
    verify_token: Optional[str] = None
    app_secret: Optional[str] = None
    api_version: SUPPORTED_API_VERS = 'v7.0'

    page: Optional[Me] = None

    def __init__(self, *,
                 auto_mark_as_seen: bool = True,
                 access_token: Optional[str] = None,
                 verify_token: Optional[str] = None,
                 app_secret: Optional[str] = None,
                 persistent_menu: Optional[List[MenuPerLocale]] = None,
                 greeting: Optional[List[GreetingPerLocale]] = None,
                 whitelisted_domains: Optional[List[HttpUrl]] = None,
                 skip_quick_reply: bool = True,
                 prevent_repeated_reply: bool = True,
                 ):
        # Non-Dynamic Change
        self.prevent_repeated_reply = prevent_repeated_reply
        if prevent_repeated_reply:
            self.reply_cache = Cache(maxsize=10000, ttl=60 * 15, default=None)

        self.skip_quick_reply = skip_quick_reply
        self.auto_mark_as_seen = auto_mark_as_seen

        if access_token:
            self.access_token = access_token
        else:
            logger.error("access_token is required.")
            raise Exception("access_token is required.")

        self.verify_token = verify_token
        self.app_secret = app_secret


        if persistent_menu or greeting or whitelisted_domains:
            profile = MessengerProfile()

            if persistent_menu:
                profile.persistent_menu = persistent_menu
            if greeting:
                profile.greeting = greeting
            if whitelisted_domains:
                profile.whitelisted_domains = whitelisted_domains

            self._set_profile_property_sync(profile)

        # self._after_send = options.pop('after_send', None)
        # self._api_ver = options.pop('api_ver', 'v7.0')
        # if self._api_ver not in SUPPORTED_API_VERS:
        #     raise ValueError('Unsupported API Version : ' + self._api_ver)

    _entryCaster = {
        MessageEntry: MessageEvent,
        PostbackEntry: PostBackEvent,
        ReferralEntry: ReferralEvent,
        DeliveriesEntry: DeliveriesEvent
    }

    # these are set by decorators or the 'set_webhook_handler' method
    _webhook_handlers = {}
    _webhook_handlers_sync = {}

    _quick_reply_callbacks = {}
    _button_callbacks = {}
    _delivered_callbacks = {}

    _quick_reply_callbacks_key_regex = {}
    _button_callbacks_key_regex = {}
    _delivered_callbacks_key_regex = {}

    _after_send = None

    @property
    def base_api_furl(self) -> furl:
        furl_url = furl("https://graph.facebook.com/") / self.api_version
        # furl_url.args['access_token'] = self.access_token
        return furl_url

    def add_verification_middleware(self, app: FastAPI):
        from songmam import VerificationMiddleware
        app.add_middleware(VerificationMiddleware, verify_token=self.verify_token)

    def handle_webhook_sync(self, webhook: Webhook):
        for entry in webhook.entry:
            handler = self._webhook_handlers_sync.get(type(entry.theMessaging))
            if handler:
                handler(MessageEvent(entry))
            else:
                logger.warning("there's no {} handler", type(entry.theMessaging))

    async def handle_webhook(self, webhook: Webhook, request: Request):

        # TODO: Convert this to middleware
        # Do the Webhook validation
        # https://developers.facebook.com/docs/messenger-platform/webhook#security
        if self.app_secret:
            header_signature = request.headers['X-Hub-Signature']
            if len(header_signature) == 45 and header_signature.startswith('sha1='):
                header_signature = header_signature[5:]
            else:
                raise NotImplementedError("Dev: how to handle this?")

            body = await request.body()
            expected_signature = hmac.new(str.encode(self.app_secret), body, hashlib.sha1).hexdigest()

            if expected_signature != header_signature:
                raise AssertionError('SIGNATURE VERIFICATION FAIL')

        for entry in webhook.entry:
            entry_type = type(entry)
            handler = self._webhook_handlers.get(entry_type)
            eventConstructor = self._entryCaster.get(entry_type)
            if handler:
                event = eventConstructor(entry)

                if entry_type is MessageEntry:
                    if self.auto_mark_as_seen:
                        self.mark_seen_sync(event.sender)

                    if event.is_quick_reply:
                        matched_callbacks = self.get_quick_reply_callbacks(event)
                        for callback in matched_callbacks:
                            await callback(event)

                elif entry_type is PostbackEntry:
                    matched_callbacks = self.get_postback_callbacks(event)
                    for callback in matched_callbacks:
                        await callback(event)
                elif entry_type is ReferralEntry:
                    pass

                elif entry_type is DeliveriesEntry:
                    pass

                await handler(event)
            else:
                logger.warning("there's no {} handler", type(entry.theMessaging))

    @property
    def id(self):
        if self.page is None:
            self._fetch_page_info_sync()

        return self.page.id

    @property
    def name(self):
        if self.page is None:
            self._fetch_page_info_sync()

        return self.page.name

    def _fetch_page_info_sync(self):
        r = requests.get(self.base_api_furl / "me",
                         params={"access_token": self.access_token},
                         headers={'Content-type': 'application/json'})

        if r.status_code != requests.codes.ok:
            print(r.text)
            return

        self.page = Me.parse_raw(r.text)

    async def _fetch_page_info(self):
        async with httpx.AsyncClient(base_url=self.base_api_furl.url, headers={'Content-type': 'application/json'},
                                     params={"access_token": self.access_token}) as client:
            response = await client.get(
                f"/me"
            )

        if response.status_code != 200:
            raise Exception(response.text)

        self.page = Me.parse_raw(response.text)

    def get_user_profile_sync(self, fb_user_id) -> UserProfile:
        r = requests.get(self.base_api_furl / fb_user_id,
                         params={"access_token": self.access_token},
                         headers={'Content-type': 'application/json'})

        if r.status_code != requests.codes.ok:
            raise ConnectionError(r.text)

        user_profile = UserProfile.parse_raw(r.text)
        return user_profile

    async def get_user_profile(self, user: Type[ThingWithId]) -> UserProfile:
        async with httpx.AsyncClient(base_url=self.base_api_furl.url, headers={'Content-type': 'application/json'},
                                     params={"access_token": self.access_token}) as client:
            response = await client.get(
                f"/{user.id}"
            )

        if response.status_code != 200:
            raise Exception(response.text)

        user_profile = UserProfile.parse_raw(response.text)
        return user_profile

    def get_messenger_code(self, ref=None, image_size=1000):
        d = {}
        d['type'] = 'standard'
        d['image_size'] = image_size
        if ref:
            d['data'] = {'ref': ref}

        r = requests.post(self._api_uri("me/messenger_codes"),
                          params={"access_token": self.access_token},
                          json=d,
                          headers={'Content-type': 'application/json'})
        if r.status_code != requests.codes.ok:
            print(r.text)
            return None

        data = json.loads(r.text)
        if 'uri' not in data:
            raise ValueError('Could not fetch messener code : GET /' +
                             self.api_version + '/me')

        return data['uri']

    def send_native_sync(self, payload: Union[CompletePayload], callback_sync=None) -> SendResponse:
        f_url = self.base_api_furl / "me/messages"
        data = payload.json(exclude_none=True)
        response = requests.post(f_url.url,
                                 params={"access_token": self.access_token},
                                 data=data,
                                 headers={'Content-type': 'application/json'})

        if response.status_code != requests.codes.ok:
            print(response.text)

        if callback_sync is not None:
            callback_sync(payload, response)

        if self._after_send is not None:
            self._after_send(payload, response)

        return SendResponse.parse_raw(response.text)

    async def send_native(self, payload: Union[CompletePayload], callback=None) -> SendResponse:

        data = payload.json(exclude_none=True)

        async with httpx.AsyncClient(base_url=self.base_api_furl.url, headers={'Content-type': 'application/json'},
                                     params={"access_token": self.access_token}) as client:
            response = await client.post(
                "/me/messages",
                data=data,
            )

        if response.status_code != 200:
            raise Exception(response.text)

        if callback is not None:
            callback_output = callback(payload, response)
            if callback_output is Awaitable:
                await callback

        if self._after_send is not None:
            self._after_send(payload, response)

        return SendResponse.parse_raw(response.text)

    def send_sync(self, *args, **kwargs):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        result = loop.run_until_complete(self.send(*args, **kwargs))

        return result

    async def send(self,
                   recipient: Union[Sender, str],
                   message: Optional[str] = None,
                   *,
                   buttons: Optional[List[AllButtonTypes]] = None,
                   quick_replies: Optional[List[QuickReply]] = None,
                   generic_elements: Optional[List[GenericElement]] = None,
                   image_aspect_ratio: Optional[Literal["horizontal", "square"]] = None,
                   media_element: Optional[MediaElement] = None,
                   media_sharable: Optional[bool] = None,
                   messaging_type: Optional[MessagingType] = MessagingType.RESPONSE,
                   tag: Optional[MessageTag] = None,
                   notification_type: Optional[NotificationType] = NotificationType.REGULAR,
                   callback: Optional[callable] = None
                   ):
        # auto cast
        if isinstance(recipient, str):
            recipient = Sender(id=recipient)

        # auto cast 2
        if buttons:
            payload = CompletePayload(
                recipient=recipient,
                message=Message(
                    attachment=TemplateAttachment(
                        payload=PayloadButtonTemplate(
                            template_type='button',
                            text=message,
                            buttons=buttons
                        )
                    ),
                    quick_replies=quick_replies
                ),
                messaging_type=messaging_type,
                tag=tag,
                notification_type=notification_type,
            )
        elif generic_elements:
            payload = CompletePayload(
                recipient=recipient,
                message=Message(
                    attachment=TemplateAttachment(
                        payload=PayloadGeneric(
                            elements=generic_elements,
                            image_aspect_ratio=image_aspect_ratio
                        )
                    ),
                    quick_replies=quick_replies
                ),
                messaging_type=messaging_type,
                tag=tag,
                notification_type=notification_type,
            )
        elif media_element:
            payload = CompletePayload(
                recipient=recipient,
                message=Message(
                    attachment=TemplateAttachment(
                        payload=PayloadMedia(
                            elements=[media_element],
                            sharable=media_sharable,
                        )
                    ),
                    quick_replies=quick_replies
                ),
                messaging_type=messaging_type,
                tag=tag,
                notification_type=notification_type,
            )
        else:
            payload = CompletePayload(
                recipient=recipient,
                message=Message(
                    text=message,
                    quick_replies=quick_replies
                ),
                messaging_type=messaging_type,
                tag=tag,
                notification_type=notification_type,
            )

        return await self.send_native(payload, callback=callback)

    # def reply_sync(self, message_to_reply_to: MessageEvent, message: ContentButton, *, quick_replies=None,
    #                metadata=None,
    #                notification_type=None, tag: Optional[MessageTag] = None, callback: Optional[callable] = None):
    #
    #     if self.prevent_repeated_reply:
    #         message_id = message_to_reply_to.entry.theMessaging.message.mid
    #         if message_id not in self.reply_cache:
    #             # good to go
    #             self.reply_cache.set(message_id, True)
    #         else:
    #             logger.warning("Songmum prevented a message from being reply to the same event multiple times.")
    #             logger.warning(message_to_reply_to)
    #             return
    #
    #     return self.send_native_sync(
    #         CompletePayload(
    #             recipient=message_to_reply_to.sender,
    #             message=message.message
    #         ), callback_sync=callback)
    #
    # async def reply(self, message_to_reply_to: MessageEvent, message, *, quick_replies=None, metadata=None,
    #                 notification_type=None, tag: Optional[MessageTag] = None, callback: Optional[callable] = None):
    #
    #     if self.prevent_repeated_reply:
    #         message_id = message_to_reply_to.entry.theMessaging.message.mid
    #         if message_id not in self.reply_cache:
    #             # good to go
    #             self.reply_cache.set(message_id, True)
    #         else:
    #             logger.warning("Songmum prevented a message from being reply to the same event multiple times.")
    #             logger.warning(message_to_reply_to)
    #             return
    #
    #     return await self.send_native(
    #         CompletePayload(
    #             recipient=message_to_reply_to.sender,
    #             message=message.message
    #         ),
    #         callback=callback
    #     )

    def typing_on_sync(self, recipient: Type[ThingWithId]):
        payload = SenderActionPayload(recipient=recipient,
                                      sender_action=SenderAction.TYPING_ON)

        self.send_native_sync(payload)

    async def typing_on(self, recipient: Type[ThingWithId]):
        payload = SenderActionPayload(recipient=recipient,
                                      sender_action=SenderAction.TYPING_ON)

        return await self.send_native(payload)

    def typing_off_sync(self, recipient: Type[ThingWithId]):
        payload = SenderActionPayload(recipient=recipient,
                                      sender_action=SenderAction.TYPING_OFF)

        self.send_native_sync(payload)

    async def typing_off(self, recipient: Type[ThingWithId]):
        payload = SenderActionPayload(recipient=recipient,
                                      sender_action=SenderAction.TYPING_OFF)

        return await self.send_native(payload)

    def mark_seen_sync(self, recipient: Type[ThingWithId]):
        payload = SenderActionPayload(recipient=recipient,
                                      sender_action=SenderAction.MARK_SEEN)

        self.send_native_sync(payload)

    async def mark_seen(self, recipient: Type[ThingWithId]):
        payload = SenderActionPayload(recipient=recipient,
                                      sender_action=SenderAction.MARK_SEEN)

        return await self.send_native(payload)

    """
    messenger profile (see https://developers.facebook.com/docs/messenger-platform/reference/messenger-profile-api)
    """

    def _set_profile_property_sync(self, data: MessengerProfile):

        f_url = self.base_api_furl / "me" / "messenger_profile"
        r = requests.post(f_url.url,
                          params={"access_token": self.access_token},
                          data=data.json(exclude_none=True),
                          headers={'Content-type': 'application/json'})

        if r.status_code != requests.codes.ok:
            raise Exception(r.text)

    def _del_profile_property_sync(self, properties: Set[MessengerProfileProperty]):
        f_url = self.base_api_furl / "me" / "messenger_profile"
        r = requests.delete(f_url.url,
                            params={"access_token": self.access_token},
                            data=json.dumps({
                                'fields': [p.value for p in properties]
                            }),
                            headers={'Content-type': 'application/json'})

        if r.status_code != requests.codes.ok:
            logger.error("Facebook Server replied" + r.text)
            raise Exception(r.text)

    """
    Custom User Settings
    """

    def get_user_settings(self, user_id: str):
        f_url = self.base_api_furl / 'me' / 'custom_user_settings'
        params = {
            "access_token": self.access_token,
            "psid": user_id
        }
        r = requests.get(f_url.url,
                         params=params)

        if r.status_code != requests.codes.ok:
            raise Exception(r.text)

        # TODO: create object for this GET Request https://developers.facebook.com/docs/messenger-platform/send-messages/persistent-menu
        return r.json()

    def set_user_menu(self, sender: Type[ThingWithId], menus: List[MenuPerLocale]):
        if isinstance(menus, MenuPerLocale):
            menus = [menus]

        f_url = self.base_api_furl / 'me' / 'custom_user_settings'
        r = requests.post(f_url.url,
                          params={"access_token": self.access_token},
                          data=UserPersistentMenu(
                              psid=sender.id,
                              persistent_menu=menus
                          ).json(),
                          headers={'Content-type': 'application/json'})

        if r.status_code != requests.codes.ok:
            raise Exception(r.text)

    def _set_user_menu(self, payload: UserPersistentMenu):
        f_url = self.base_api_furl / 'me' / 'custom_user_settings'
        r = requests.post(f_url.url,
                          params={"access_token": self.access_token},
                          data=payload.json(),
                          headers={'Content-type': 'application/json'})

        if r.status_code != requests.codes.ok:
            raise Exception(r.text)

    def delete_user_menu(self, user_id: str):
        f_url = self.base_api_furl / 'me' / 'custom_user_settings'

        params = {
            "access_token": self.access_token,
            "psid": user_id,
            "params": "[%22persistent_menu%22]"
        }
        r = requests.delete(f_url.url,
                            params=params)

        if r.status_code != requests.codes.ok:
            raise Exception(r.text)

    async def create_persona(self, persona: Persona) -> PersonaResponse:
        data = persona.json()

        async with httpx.AsyncClient(base_url=self.base_api_furl.url, headers={'Content-type': 'application/json'},
                                     params={"access_token": self.access_token}) as client:
            response = await client.post(
                "/me/personas",
                data=data,
            )

        if response.status_code != 200:
            raise Exception(response.text)

        return PersonaResponse.parse_raw(response.text)

    async def get_persona(self, id):

        async with httpx.AsyncClient(base_url=self.base_api_furl.url, headers={'Content-type': 'application/json'},
                                     params={"access_token": self.access_token}) as client:
            response = await client.get(
                f"/{id}",
            )

        if response.status_code != 200:
            raise Exception(response.text)

        return PersonaWithId.parse_raw(response.text)

    async def get_all_personas(self) -> List[PersonaWithId]:

        async with httpx.AsyncClient(base_url=self.base_api_furl.url, headers={'Content-type': 'application/json'},
                                     params={"access_token": self.access_token}) as client:
            response = await client.get(
                f"/me/personas",
            )

        if response.status_code != 200:
            raise Exception(response.text)

        response = AllPerosnasResponse.parse_raw(response.text)

        # There might be a need to implement paging in future
        # Note: https://developers.facebook.com/docs/graph-api/using-graph-api/#cursors

        return response.data

    async def delete_persona(self, id):

        async with httpx.AsyncClient(base_url=self.base_api_furl.url, headers={'Content-type': 'application/json'},
                                     params={"access_token": self.access_token}) as client:
            response = await client.delete(
                f"/{id}",
            )

        if response.status_code != 200:
            raise Exception(response.text)

        return PersonaDeleteResponse.parse_raw(response.text)

    """
    handlers and decorations
    """

    # def set_webhook_handler(self, scope, callback):
    #     """
    #     Allows adding a webhook_handler as an alternative to the decorators
    #     """
    #     scope = scope.lower()
    #
    #     if scope == 'after_send':
    #         self._after_send = callback
    #         return
    #
    #     if scope not in Page.WEBHOOK_ENDPOINTS:
    #         raise ValueError("The 'scope' argument must be one of {}.".format(Page.WEBHOOK_ENDPOINTS))
    #
    #     self._webhook_handlers[scope] = callback

    def handle_optin(self, func):
        self._webhook_handlers['optin'] = func

    def handle_message_sync(self, func: callable):
        self._webhook_handlers_sync[MessageEntry] = func

    def handle_message(self, func: callable):
        self._webhook_handlers[MessageEntry] = func

    def handle_echo_sync(self, func: callable):
        self._webhook_handlers_sync[EchoEntry] = func

    def handle_echo(self, func: callable):
        self._webhook_handlers[EchoEntry] = func

    def handle_delivery(self, func):
        self._webhook_handlers[DeliveriesEntry] = func

    def handle_postback_sync(self, func):
        self._webhook_handlers_sync[PostbackEntry] = func

    def handle_postback(self, func):
        self._webhook_handlers[PostbackEntry] = func

    # def handle_read(self, func):
    #     self._webhook_handlers['read'] = func
    #
    # def handle_account_linking(self, func):
    #     self._webhook_handlers['account_linking'] = func

    # def handle_referral_sync(self, func):
    #     self._webhook_handlers_sync['referral'] = func

    def handle_referral(self, func):
        self._webhook_handlers[ReferralEntry] = func

    #
    # def handle_game_play(self, func):
    #     self._webhook_handlers['game_play'] = func
    #
    # def handle_pass_thread_control(self, func):
    #     self._webhook_handlers['pass_thread_control'] = func
    #
    # def handle_take_thread_control(self, func):
    #     self._webhook_handlers['take_thread_control'] = func
    #
    # def handle_request_thread_control(self, func):
    #     self._webhook_handlers['request_thread_control'] = func
    #
    # def handle_app_roles(self, func):
    #     self._webhook_handlers['app_roles'] = func
    #
    # def handle_policy_enforcement(self, func):
    #     self._webhook_handlers['policy_enforcement'] = func
    #
    # def handle_checkout_update(self, func):
    #     self._webhook_handlers['checkout_update'] = func
    #
    # def handle_payment(self, func):
    #     self._webhook_handlers['payment'] = func
    #
    # def handle_standby(self, func):
    #     self._webhook_handlers['standby'] = func

    def after_send(self, func):
        self._after_send = func

    def callback(self, payloads=None, quick_reply=True, button=True):

        def wrapper(func):
            if payloads is None:
                return func

            for payload in payloads:
                if quick_reply:
                    self._quick_reply_callbacks[payload] = func
                if button:
                    self._button_callbacks[payload] = func

            return func

        return wrapper

    def get_quick_reply_callbacks(self, event: MessageEvent):
        callbacks = []
        for key in self._quick_reply_callbacks.keys():
            if key not in self._quick_reply_callbacks_key_regex:
                self._quick_reply_callbacks_key_regex[key] = re.compile(key + '$')

            if self._quick_reply_callbacks_key_regex[key].match(event.quick_reply.payload):
                callbacks.append(self._quick_reply_callbacks[key])

        return callbacks

    def get_postback_callbacks(self, event: PostBackEvent):
        callbacks = []
        for key in self._button_callbacks.keys():
            if key not in self._button_callbacks_key_regex:
                self._button_callbacks_key_regex[key] = re.compile(key + '$')

            if self._button_callbacks_key_regex[key].match(event.payload):
                callbacks.append(self._button_callbacks[key])

        return callbacks



from typing import Optional, List, Union

from pydantic import BaseModel, validator, conlist

from songmam.facebook.entries.message.attachment import Attachment
from songmam.facebook.entries.base import MessagingWithTimestamp
from songmam.facebook import ThingWithId


class Sender(ThingWithId):
    user_ref: Optional[str]


class QuickReply(BaseModel):
    """A quick_reply payload is only provided with a text message when the user tap on a Quick Replies button."""
    payload: str


class ReplyTo(BaseModel):
    """"""
    mid: str  # Reference to the message ID that this message is replying to


class Message(BaseModel):
    mid: str  # Message ID
    text: Optional[str] = None  # Text of message
    quick_reply: Optional[QuickReply] = None
    reply_to: Optional[ReplyTo] = None
    attachments: Optional[List[Attachment]] = None


class Postback(BaseModel):
    title: str
    payload: str


class Messaging(MessagingWithTimestamp):
    sender: Sender
    message: Message


class MessageEntry(BaseModel):
    id: str
    time: int
    messaging: conlist(Messaging, min_items=1, max_items=1)

    @property
    def theMessaging(self):
        return self.messaging[0]



from pydantic import BaseModel

from songmam.facebook import ThingWithId


class BaseMessaging(BaseModel):
    sender: ThingWithId
    recipient: ThingWithId


class MessagingWithTimestamp(BaseMessaging):
    timestamp: int






from typing import Optional, Literal

from pydantic import BaseModel

from songmam.facebook.entries.base import MessagingWithTimestamp

class Referral(BaseModel):
    ref: str
    source: Literal["MESSENGER_CODE", "DISCOVER_TAB", "ADS", "SHORTLINK", "CUSTOMER_CHAT_PLUGIN"]
    type: str
    ad_id: Optional[str]
    referer_uri: Optional[str]


class ReferralEntry(MessagingWithTimestamp):
    referral: Referral

# {
#     "recipient": {
#         "id": "<PSID>"
#     },
#     "recipient": {
#         "id": "<PAGE_ID>"
#     },
#     "timestamp": 1458692752478,
#     "referral": {
#                     "ref": < REF_DATA_PASSED_IN_M.ME_PARAM >,
#     "source": "SHORTLINK",
#     "type": "OPEN_THREAD",
# }
# }


from enum import auto

from songmam.utils import AutoNameLower


class SenderAction(AutoNameLower):
    """https://developers.facebook.com/docs/messenger-platform/send-messages/sender-actions"""
    TYPING_ON = auto()
    TYPING_OFF = auto()
    MARK_SEEN = auto()


from typing import List, Union, Optional

from loguru import logger
from pydantic import BaseModel, validator

from songmam.facebook.entries.deliveries import DeliveriesEntry
from songmam.facebook.entries.echo import EchoEntry
from songmam.facebook.entries.messages import MessageEntry
from songmam.facebook.entries.postback import PostbackEntry


class Webhook(BaseModel):
    """An object contains one or more entries
    https://developers.facebook.com/docs/messenger-platform/reference/webhook-events/#payload
    """
    object: str
    entry: List[Union[DeliveriesEntry, PostbackEntry, MessageEntry, EchoEntry]]

    @validator('object')
    def object_equal_page(cls, v):
        if v != 'page':
            error_msg = "only support page subscription"
            logger.error(error_msg)
            raise ValueError(error_msg)
        return v



"""
Payload is final data  form ready to use with request api
"""
from enum import auto
from typing import List, Union, Optional, Any

from pydantic import BaseModel, validator, conlist

from songmam.facebook import ThingWithId
from songmam.facebook.entries.messages import Sender
from songmam.facebook.messaging.message_tags import MessageTag
from songmam.facebook.messaging.messaging_type import MessagingType
from songmam.facebook.messaging.notification_type import NotificationType
from songmam.facebook.messaging.sender_action import SenderAction
from songmam.facebook.messaging.templates.button import AllButtonTypes
from songmam.facebook.messaging.templates import Message
from songmam.facebook.send import SendRecipient


class CompletePayload(BaseModel):
    recipient: Union[SendRecipient, Sender, ThingWithId]
    template_type: Optional[str] = None
    message: Optional[Message]
    messaging_type: Optional[MessagingType] = MessagingType.RESPONSE
    tag: Optional[MessageTag]
    notification_type: Optional[NotificationType] = NotificationType.REGULAR

    persona_id: Optional[str] = None


class SenderActionPayload(CompletePayload):
    sender_action: SenderAction

class SendingQuickRepliesEntry(CompletePayload):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/buttons/quick-replies
    """
    message: Any
    buttons: conlist(AllButtonTypes, max_items=3, min_items=1)  # Set of 1-3 buttons that appear as call-to-actions.



from songmam.utils import AutoNameLower


class SenderAction(AutoNameLower):
    """https://developers.facebook.com/docs/messenger-platform/send-messages/sender-actions"""
    TYPING_ON = auto()
    TYPING_OFF = auto()
    MARK_SEEN = auto()





from typing import Literal, Optional, List

from pydantic import BaseModel

from songmam.facebook.entries.base import MessagingWithTimestamp
from songmam.facebook.entries.echo import Message
from songmam.facebook.entries.messages import MessageEntry, Messaging


class PostbackReferral(BaseModel):
    ref: str
    source: Literal["SHORTLINK", "ADS"]
    type: Literal["OPEN_THREAD"]

class Postback(BaseModel):
    title: str
    payload: str
    referral: Optional[PostbackReferral]

class PostbackMessaging(Messaging):
    message: Optional[Message]
    postback: Postback

class PostbackEntry(MessageEntry):
    # message: Optional[Message]
    messaging: List[PostbackMessaging]
    # postback: Optional[Postback]

# {
#   "recipient":{
#     "id":"<PSID>"
#   },
#   "recipient":{
#     "id":"<PAGE_ID>"
#   },
#   "timestamp":1458692752478,
#   "postback":{
#     "title": "<TITLE_FOR_THE_CTA>",
#     "payload": "<USER_DEFINED_PAYLOAD>",
#     "referral": {
#       "ref": "<USER_DEFINED_REFERRAL_PARAM>",
#       "source": "<SHORTLINK>",
#       "type": "OPEN_THREAD",
#     }
#   }
# }

"""
{"object":"page","entry":[{"id":"103157244728633","time":1595061969020,"messaging":[{"recipient":{"id":"2892682217518683"},"recipient":{"id":"103157244728633"},"timestamp":1595061968847,"postback":{"title":"menu 2","payload":"menu 2"}}]}]}
"""


from enum import auto

from songmam.utils import AutoName


class MessageTag(AutoName):
    """
    https://developers.facebook.com/docs/messenger-platform/send-messages/message-tags
    """
    CONFIRMED_EVENT_UPDATE = auto()
    POST_PURCHASE_UPDATE = auto()
    ACCOUNT_UPDATE = auto()
    HUMAN_AGENT = auto()


from typing import Literal

from pydantic import BaseModel

from songmam.facebook.entries.base import MessagingWithTimestamp


class Reaction(BaseModel):
    reaction: Literal['smile', 'angry', 'sad', 'wow', 'love', 'like', 'dislike', 'other']
    emoji: str
    action: str
    mid: str

class ReactionEntry(MessagingWithTimestamp):
    reaction: Reaction



# {
#    "recipient":{
#       "id":"<PSID>"
#    },
#    "recipient":{
#       "id":"<PAGE_ID>"
#    },
#    "timestamp":1458668856463,
#    "reaction":{
#          "reaction": "smile|angry|sad|wow|love|like|dislike|other",
#          "emoji": "\u{2764}\u{FE0F}",
#          "action": "react|unreact",
#          "mid": "<MID_OF_ReactedTo_Message>",
#    }
# }


"""
    https://developers.facebook.com/docs/messenger-platform/reference/messenger-profile-api/ice-breakers
    """
from pydantic import BaseModel


class IceBreaker(BaseModel):
    question: str
    payload: str



from pydantic import BaseModel

from songmam.facebook.entries.base import MessagingWithTimestamp


class GamePlay(BaseModel):
    game_id: str
    player_id: str
    context_type: str
    context_id: str
    score: int
    payload: str

class GamePlayEntries(MessagingWithTimestamp):
    game_play: GamePlay

# {
#   "recipient": {
#     "id": "<PSID>"
#   },
#   "recipient": {
#     "id": "<PAGE_ID>"
#   },
#   "timestamp": 1469111400000,
#   "game_play": {
#     "game_id": "<GAME-APP-ID>",
#     "player_id": "<PLAYER-ID>",
#     "context_type": "<CONTEXT-TYPE:SOLO|THREAD>",
#     "context_id": "<CONTEXT-ID>", # If a Messenger Thread context
#     "score": <SCORE-NUM>, # If a classic score based game
#     "payload": "<PAYLOAD>" # If a rich game
#   }
# }


from enum import auto

from pydantic import BaseModel

from songmam.utils import AutoName


class Locale(AutoName):
    """https://developers.facebook.com/docs/messenger-platform/messenger-profile/supported-locales"""
    default = auto()
    en_US = auto()
    ca_ES = auto()
    cs_CZ = auto()
    cx_PH = auto()
    cy_GB = auto()
    da_DK = auto()
    de_DE = auto()
    eu_ES = auto()
    en_UD = auto()
    es_LA = auto()
    es_ES = auto()
    gn_PY = auto()
    fi_FI = auto()
    fr_FR = auto()
    gl_ES = auto()
    hu_HU = auto()
    it_IT = auto()
    ja_JP = auto()
    ko_KR = auto()
    nb_NO = auto()
    nn_NO = auto()
    nl_NL = auto()
    fy_NL = auto()
    pl_PL = auto()
    pt_BR = auto()
    pt_PT = auto()
    ro_RO = auto()
    ru_RU = auto()
    sk_SK = auto()
    sl_SI = auto()
    sv_SE = auto()
    th_TH = auto()
    tr_TR = auto()
    ku_TR = auto()
    zh_CN = auto()
    zh_HK = auto()
    zh_TW = auto()
    af_ZA = auto()
    sq_AL = auto()
    hy_AM = auto()
    az_AZ = auto()
    be_BY = auto()
    bn_IN = auto()
    bs_BA = auto()
    bg_BG = auto()
    hr_HR = auto()
    nl_BE = auto()
    en_GB = auto()
    et_EE = auto()
    fo_FO = auto()
    fr_CA = auto()
    ka_GE = auto()
    el_GR = auto()
    gu_IN = auto()
    hi_IN = auto()
    is_IS = auto()
    id_ID = auto()
    ga_IE = auto()
    jv_ID = auto()
    kn_IN = auto()
    kk_KZ = auto()
    lv_LV = auto()
    lt_LT = auto()
    mk_MK = auto()
    mg_MG = auto()
    ms_MY = auto()
    mt_MT = auto()
    mr_IN = auto()
    mn_MN = auto()
    ne_NP = auto()
    pa_IN = auto()
    sr_RS = auto()
    so_SO = auto()
    sw_KE = auto()
    tl_PH = auto()
    ta_IN = auto()
    te_IN = auto()
    ml_IN = auto()
    uk_UA = auto()
    uz_UZ = auto()
    vi_VN = auto()
    km_KH = auto()
    tg_TJ = auto()
    ar_AR = auto()
    he_IL = auto()
    ur_PK = auto()
    fa_IR = auto()
    ps_AF = auto()
    my_MM = auto()
    qz_MM = auto()
    or_IN = auto()
    si_LK = auto()
    rw_RW = auto()
    cb_IQ = auto()
    ha_NG = auto()
    ja_KS = auto()
    br_FR = auto()
    tz_MA = auto()
    co_FR = auto()
    as_IN = auto()
    ff_NG = auto()
    sc_IT = auto()
    sz_PL = auto()

class ThingWithLocale(BaseModel):
    locale: Locale = Locale.default



from typing import Optional, List, Literal

from pydantic import BaseModel


class AuxiliaryFields(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-boarding-pass#auxiliary_field
    """
    label: str
    value: str


class SecondaryFields(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-boarding-pass#secondary_field
    """
    label: str
    value: str


class DepartureAirport(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-itinerary#departure_airport
    """
    airport_code: str
    city: str
    terminal: str
    gate: str


class ArrivalAirport(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-itinerary#arrival_airport
    """
    airport_code: str
    city: str


class FlightSchedule(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-itinerary#flight_schedule
    """
    boarding_time: Optional[str]
    departure_time: str
    arrival_time: Optional[str]


class FlightInfo(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-itinerary#flight_info
    """
    flight_number: str
    departure_airport: DepartureAirport
    arrival_airport: ArrivalAirport
    flight_schedule: FlightSchedule


class BoardingPass(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-boarding-pass#boarding_pass
    """
    passenger_name: str
    pnr_number: str
    travel_class: Optional[str]
    seat: Optional[str]
    auxiliary_fields: Optional[List[AuxiliaryFields]]
    secondary_fields: Optional[List[SecondaryFields]]
    logo_image_url: str
    header_image_url: Optional[str]
    header_text_field: Optional[str]  # in ref is "field" /// Optional. Text for the header field.
    qr_code: Optional[str]  # Not available if barcode_image_urlis used.
    barcode_image_url: Optional[str]  # Not available if qr_code is used.
    above_bar_code_image_url: Optional[str]
    flight_info: FlightInfo


class PayloadAirlineBoardingPass(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-boarding-pass
    """
    template_type: Literal["airline_boardingpass"]
    intro_message: str
    locale: str
    theme_color: Optional[str]
    boarding_pass: List[BoardingPass]


class PayloadAirlineCheckin(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-checkin
    """
    template_type: Literal["airline_checkin"]
    intro_message: str
    locale: str
    pnr_number: Optional[str]
    checkin_url: str
    flight_info: List[FlightInfo]


class PassengerInfo(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-itinerary#passenger_info
    """
    passenger_id: str
    ticket_number: Optional[str]
    name: str


class ProductInfo(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-itinerary#product_info
    """
    title: str
    value: str


class PassengerSegmentInfo(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-itinerary#passenger_segment_info
    """
    segment_id: str
    passenger_id: PassengerInfo.passenger_id
    seat: str
    seat_type: str
    product_info: Optional[
        List[ProductInfo]]  # List of products the passenger purchased. Maximum of 4 items is supported.


class PriceInfo(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-itinerary#price_info
    """
    title: str
    amount: int
    currency: Optional[str]


class PayloadAirlineItinerary(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-itinerary
    """
    template_type: Literal["airline_itinerary"]
    intro_message: str
    locale: str
    theme_color: Optional[str]
    pnr_number: str
    passenger_info: List[PassengerInfo]
    flight_info: List[FlightInfo]
    passenger_segment_info: List[PassengerSegmentInfo]
    price_info: Optional[List[PriceInfo]]
    base_price: Optional[int]
    tax: Optional[int]
    total_price: int
    currency: str


class UpdateFlightInfo(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-flight-update#update_flight_info
    """
    flight_number: str
    departure_airport: DepartureAirport
    arrival_airport: ArrivalAirport
    flight_schedule: FlightSchedule


class PayloadAirlineUpdate(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/airline-flight-update#payload
    """
    template_type: Literal["airline_update"]
    intro_message: str
    theme_color: Optional[str]
    update_type: Literal["delay", "gate_change", "cancellation"]
    locale: str
    pnr_number: Optional[str]
    update_flight_info: UpdateFlightInfo


from dataclasses import dataclass
from typing import Optional, List

from songmam.facebook.messaging.quick_replies import QuickReply
from songmam.facebook.messaging.templates import TemplateAttachment, Message
from songmam.facebook.messaging.templates.receipt import ReceiptElements, Address, Summary, Adjustments, PayloadReceipt


@dataclass
class ContentReceipt:
    quick_replies: Optional[List[QuickReply]]
    sharable: Optional[bool]
    recipient_name: str
    merchant_name: Optional[str]
    order_number: str
    currency: str
    payment_method: str  # This can be a custom string, such as, "Visa 1234".
    timestamp: Optional[str]
    elements: Optional[List[ReceiptElements]]
    address: Optional[Address]
    summary: Summary
    adjustments: Optional[List[Adjustments]]

    @property
    def message(self):
        message = Message()

        if self.elements:
            payload = PayloadReceipt(
                template_type="receipt",
                recipient_name=self.recipient_name,
                order_number=self.order_number,
                currency=self.currency,
                payment_method=self.payment_method,  # This can be a custom string, such as, "Visa 1234".
                summary=self.summary,
            )
            payload.sharable = self.sharable
            payload.merchant_name = self.merchant_name
            payload.timestamp = self.timestamp
            payload.elements = self.elements
            payload.address = self.address
            payload.adjustments = self.adjustments
            message.attachment = TemplateAttachment(
                payload=payload
            )
        if self.quick_replies:
            message.quick_replies = self.quick_replies

        return message



from typing import Optional, Literal, List

from pydantic import BaseModel, validator, conlist

from songmam.facebook.messaging.templates.button import AllButtonTypes


class MediaElement(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/media#elements
    """
    media_type: Optional[Literal["image", "video"]]
    attachment_id: Optional[str]  # Cannot be used if url is set.
    url: Optional[str]  # Cannot be used if attachment_id is set. || have to be image / vids in facebook url
    buttons: Optional[conlist(AllButtonTypes, min_items=1, max_items=3)]  # A maximum of 1 button is supported.


class PayloadMedia(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/media#payload
    """
    template_type: Literal["media"] = 'media'
    elements: conlist(MediaElement, min_items=1, max_items=1)
    sharable: Optional[bool]



from enum import auto
from typing import Literal, Optional, List, Type, Union

from pydantic import BaseModel, validator, HttpUrl, root_validator, conlist, constr

# from songmam.facebook.messaging.templates import CompletePayload as Payload_
from songmam.utils import AutoName


class BaseButton(BaseModel):
    type: str

    # make optional for default action
    title: Optional[constr(max_length=20)] = None

    def get_default_action(self):
        self.title = None
        return self

    @classmethod
    def create_default_action(cls, *args, **kargs):
        b = cls(**kargs)
        return b.get_default_action()


class URLButton(BaseButton):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/buttons/url
    Updated: 04/07/2020
    """
    type: str = 'web_url'
    url: HttpUrl
    webview_height_ratio: Literal['compact', 'tall', 'full'] = 'full'
    messenger_extensions: bool = False
    fallback_url: Optional[HttpUrl]
    webview_share_button: Optional[Literal['hide']]

    @root_validator
    def fallback_url_should_not_be_specify_if_messenger_extensions_is_false(cls, values):
        messenger_extensions, fallback_url = values.get('messenger_extensions'),values.get('fallback_url')
        if fallback_url:
            if not messenger_extensions:
                raise ValueError('`fallback_url` may only be specified if `messenger_extensions` is true.')
        return values


class PostbackButton(BaseButton):
    type: str = 'postback'
    payload: str


class CallButton(BaseButton):
    type: str = 'phone_number'
    payload: str

    # TODO: https://pypi.org/project/phonenumbers/


class LogInButton(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/send-messages/buttons#login
    """
    type: str = "account_link"
    url: HttpUrl  # Authentication callback URL. Must use HTTPS protocol.


class LogOutButton(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/buttons/logout
    """
    type: str = "account_unlink"


class GameMetadata(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/buttons/game-play#game_metadata
    """
    player_id: Optional[str]
    context_id: Optional[str]


class GamePlayButton(BaseButton):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/buttons/game-play
    """
    type: str = "game_play"
    title: str
    payload: Optional[str]
    game_metadata: Optional[GameMetadata]


# class Button(BaseModel):
#     """
#     https://developers.facebook.com/docs/messenger-platform/reference/buttons/url#properties
#     """
#     type: Literal["web_url", "postback", "phone_number", "account_link", "account_unlink", "game_play"]
#     title: str
#     payload: Optional[str]  # for type : postback / phone_number / game_play
#     url: Optional[HttpUrl]  # for type: web_url / LogIn
#     webview_height_ratio: Optional[Literal["compact", "tall", "full"]]  # for type: web_url
#     messenger_extensions: Optional[bool]  # for type: web_url
#     fallback_url: Optional[HttpUrl]  # for type: web_url
#     webview_share_button: Optional[str]  # for type: web_url
#     game_metadata: Optional[GameMetadata]  # for type : game_play

AllButtonTypes = Union[URLButton, PostbackButton, CallButton, LogInButton, LogOutButton, GamePlayButton]


# TODO: Inherit from the CompletePayload?
class PayloadButtonTemplate(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/button
    """
    template_type = 'button'
    text: str
    buttons: conlist(AllButtonTypes, min_items=1, max_items=3)  # Set of 1-3 buttons that appear as call-to-actions.

    @validator('text')
    def title_limit_to_640_characters(cls, v):
        if len(v) > 640:
            raise ValueError('UTF-8-encoded text of up to 640 characters.')
        return v


class ButtonType(AutoName):
    web_url = auto()
    postback = auto()
    phone_number = auto()
    account_link = auto()
    account_unlink = auto()
    game_play = auto()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['songmam',
 'songmam.api',
 'songmam.facebook',
 'songmam.facebook.entries',
 'songmam.facebook.entries.message',
 'songmam.facebook.messaging',
 'songmam.facebook.messaging.templates',
 'songmam.facebook.messaging.templates.airline',
 'songmam.facebook.messenger_profile']

package_data = \
{'': ['*']}

install_requires = \
['cacheout>=0.11.2,<0.12.0',
 'furl>=2.1.0,<3.0.0',
 'httpx>=0.13.3,<0.14.0',
 'loguru>=0.5.1,<0.6.0',
 'requests>=2.24.0,<3.0.0',
 'starlette>=0.13.4,<0.14.0']

setup_kwargs = {
    'name': 'songmam',
    'version': '0.5.7',
    'description': 'a fork of FBMQ',
    'long_description': None,
    'author': 'Nutchanon Ninyawee',
    'author_email': 'me@nutchanon.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from enum import auto
from typing import Optional, List

from pydantic import BaseModel, HttpUrl

from songmam.facebook.messenger_profile.get_started import GetStarted
from songmam.facebook.messenger_profile.greeting import GreetingPerLocale
from songmam.facebook.messenger_profile.ice_breakers import IceBreaker
from songmam.facebook.messenger_profile.persistent_menu import MenuPerLocale
from songmam.utils import AutoName


class MessengerProfileProperty(AutoName):
    get_started = auto()
    greeting = auto()
    ice_breakers = auto()
    persistent_menu = auto()
    whitelisted_domains = auto()
    account_linking_url = auto()

class MessengerProfile(BaseModel):
    get_started: Optional[GetStarted]
    greeting: Optional[List[GreetingPerLocale]]
    ice_breakers: Optional[List[IceBreaker]]
    persistent_menu: Optional[List[MenuPerLocale]]
    whitelisted_domains: Optional[List[HttpUrl]]
    account_linking_url: Optional[HttpUrl]




from typing import Optional, Literal, List

from pydantic import BaseModel, HttpUrl


class ReceiptElements(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/receipt#elements
    """
    title: str
    subtitle: Optional[str]
    quantity: Optional[int]
    price: float
    currency: Optional[str]
    image_url: Optional[HttpUrl]


class Address(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/receipt#address
    """
    street_1: str
    street_2: Optional[str]
    city: str
    postal_code: str
    state: str
    country: str


class Summary(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/receipt#summary
    """
    subtotal: Optional[float]
    shipping_cost: Optional[float]
    total_tax: Optional[float]
    total_cost: float


class Adjustments(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/receipt#adjustments
    """
    name: str
    amount: float


class PayloadReceipt(BaseModel):
    """
    https://developers.facebook.com/docs/messenger-platform/reference/templates/receipt
    """
    template_type: Literal["receipt"]
    sharable: Optional[bool]
    recipient_name: str
    merchant_name: Optional[str]
    order_number: str
    currency: str
    payment_method: str  # This can be a custom string, such as, "Visa 1234".
    timestamp: Optional[str]
    elements: Optional[List[ReceiptElements]]
    address: Optional[Address]
    summary: Summary
    adjustments: Optional[List[Adjustments]]



from pydantic import BaseModel

from songmam.facebook.entries.base import MessagingWithTimestamp

class Optin(BaseModel):
    ref: str
    user_ref: str

class OptinEntry(MessagingWithTimestamp):
    optin: Optin

# {
#   "recipient": {
#     "id": "<PSID>"
#   },
#   "recipient": {
#     "id": "<PAGE_ID>"
#   },
#   "timestamp": 1234567890,
#   "optin": {
#     "ref": "<PASS_THROUGH_PARAM>",
#     "user_ref": "<REF_FROM_CHECKBOX_PLUGIN>"
#   }
# }


from typing import Literal, Optional

from pydantic import BaseModel, HttpUrl


class Coordinates(BaseModel):
    lat: float
    long: float


class PayloadOfAttachment(BaseModel):
    url: HttpUrl
    title: Optional[str]
    sticker_id: Optional[int]
    coordinates: Optional[Coordinates]


class Attachment(BaseModel):
    type: Literal['audio', 'file', 'image', 'location', 'video', 'fallback', "template"]
    payload: PayloadOfAttachment